@@ -24,7 +24,9 @@
 """
 import re
 import os
+import sys
 from os import path
+import tempfile
 import json
 from collections import defaultdict
 import subprocess
@@ -33,6 +35,7 @@ from docutils import nodes
 from docutils.statemachine import string2lines, StringList
 
 from sphinx import addnodes
+from sphinx.util.console import red
 from sphinx.locale import l_, _
 from sphinx.roles import XRefRole
 from sphinx.domains import Domain, ObjType
@@ -115,33 +118,6 @@ def parse_specializer_symbol(symbol, package):
     return symbol
 
 
-def resolve_string(state_machine, package, symbol, objtype, specializer=None):
-    """
-    Resolve a symbols doc string. Will raise KeyError if the
-    symbol can't be found.
-    """
-    if objtype == "method":
-        spec = specializer[0].split(" ")[1:]
-        method_doc = METHODS[package].get(symbol, {})
-        key = tuple([parse_specializer_symbol(sym, package)
-                     for sym in spec])
-        if key not in method_doc:
-            state_machine.reporter.warning("Can't find method %s:%s specializer %s" %
-                                           (package, symbol, spec))
-        return method_doc.get(key, "")
-
-    possible_strings = DOC_STRINGS[package][symbol]
-
-    # XXX This isn't the best, the objtype is generic but the
-    # docstring will be under genericFunction because of the JSON
-    # encoder and changing the directive name doesn't seem to help
-    # either.
-    if objtype == "generic":
-        objtype = "genericFunction"
-    string = possible_strings.get(objtype, "")
-    return string
-
-
 class desc_clparameterlist(addnodes.desc_parameterlist):
     """Node for a common lisp parameter list."""
     child_text_separator = ' '
@@ -380,7 +356,7 @@ class CLsExp(ObjectDescription):
         else:
             package = ""
 
-        if name not in self.state.document.ids:
+        if type != "method" and name not in self.state.document.ids:
             signode['names'].append(name)
             signode['ids'].append(name)
             signode['first'] = (not self.names)
@@ -403,10 +379,12 @@ class CLsExp(ObjectDescription):
         name = self.names[0][1]
         description = result[1][-1]
         node = addnodes.desc_content()
+        if not package:
+            self.state_machine.reporter.warning("No package specified for symbol %s." %
+                                                name)
+            return
         try:
-            string = resolve_string(self.state_machine, package,
-                                    name, self.objtype,
-                                    self.arguments)
+            string = self.cl_doc_string()
         except KeyError:
             string = ""
             self.state_machine.reporter.warning("Can't find symbol %s:%s" %
@@ -455,6 +433,61 @@ class CLsExp(ObjectDescription):
             self.run_add_specializers(result)
         return result
 
+    def cl_doc_string(self, objtype=None):
+        """
+        Resolve a symbols doc string. Will raise KeyError if the
+        symbol can't be found.
+        """
+        package = self.env.temp_data.get('cl:package')
+        name = self.names[0][1]
+        objtype = objtype or self.objtype
+
+        possible_strings = DOC_STRINGS[package][name]
+
+        # XXX This isn't the best, the objtype is generic but the
+        # docstring will be under genericFunction because of the JSON
+        # encoder and changing the directive name doesn't seem to help
+        # either.
+        if objtype == "generic":
+            objtype = "genericFunction"
+        string = possible_strings.get(objtype, "")
+        return string
+
+
+class CLMethod(CLsExp):
+
+    option_spec = {
+        'nodoc': bool_option,
+        'noindex': bool_option,
+        'noinherit': bool_option,
+    }
+
+    def cl_doc_string(self):
+        """
+        Resolve a symbols doc string. Will raise KeyError if the
+        symbol can't be found.
+        """
+        package = self.env.temp_data.get('cl:package')
+        name = self.names[0][1]
+        objtype = self.objtype
+
+        if objtype == "method":
+            specializer = self.arguments
+            spec = specializer[0].split(" ")[1:]
+            method_doc = METHODS[package].get(name, {})
+            key = tuple([parse_specializer_symbol(sym, package)
+                         for sym in spec])
+            if key not in method_doc:
+                self.state_machine.reporter.warning("Can't find method %s:%s specializer %s, available specializers are %s" %
+                                               (package, name, key, method_doc.keys()))
+            doc = method_doc.get(key, "")
+            if doc:
+                return doc
+
+        if "noinherit" not in self.options:
+            return super(CLMethod, self).cl_doc_string("generic")
+        return ""
+
 
 class CLCurrentPackage(Directive):
     """
@@ -502,7 +535,7 @@ class CLDomain(Domain):
         'type': ObjType(l_('type'), 'type'),
         'generic': ObjType(l_('generic'), 'generic'),
         'method': ObjType(l_('method'), 'method'),
-    }
+        }
 
     directives = {
         'package': CLCurrentPackage,
@@ -511,7 +544,7 @@ class CLDomain(Domain):
         'macro': CLsExp,
         'variable': CLsExp,
         'type': CLsExp,
-        'method': CLsExp,
+        'method': CLMethod,
     }
 
     roles = {
@@ -582,6 +615,14 @@ def code_regions(text):
     return output.read()
 
 
+def save_cldomain_output(output):
+    """Save a copy of the clgit output for debugging."""
+    fd, path = tempfile.mkstemp('.log', 'cldomain-err-')
+    os.write(fd, output.encode('utf-8'))
+    os.close(fd)
+    return path
+
+
 def index_package(package, package_path, quicklisp, lisps):
     """Call an external lisp program that will return a dictionary of
     doc strings for all public symbols."""
@@ -597,7 +638,18 @@ def index_package(package, package_path, quicklisp, lisps):
     output = "\n".join([line for line in output.split("\n")
                         if not line.startswith(";")])
 
-    lisp_data = json.loads(output)
+    try:
+        lisp_data = json.loads(output)
+    except:
+        dump_path = save_cldomain_output(output)
+        error = sys.stderr
+        print >>error, red('A error occurred with the json output from cldomain\'s'
+                           ' lisp inspector,  this has been dumped to %s if you '
+                           'intend on submitting a bug please include this file '
+                           'with the sphinx error log.' % dump_path)
+
+        raise
+
     for k, v in lisp_data.items():
 
         # extract doc strings
